@@ -7,10 +7,137 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# RCS: @(#) $Id: dirlist.tcl,v 1.7 2000/08/02 07:06:52 welch Exp $
+# RCS: @(#) $Id: dirlist.tcl,v 1.11 2003/09/24 20:41:57 welch Exp $
 
-package provide httpd::dirlist 1.0
+package provide httpd::dirlist 1.1
  
+# DirList_IndexFile --
+#
+#	Define the index file for a directory
+#
+# Arguments:
+#	pat	A glob pattern for index files in a directory.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Sets the index file glob pattern.
+
+proc DirList_IndexFile {pat} {
+    global dirlist
+    set dirlist(indexpat) $pat
+}
+
+# DirList_Directory --
+#
+#	Handle a directory.  Look for the index file, falling back to
+#	the Directory Listing module, if necessary.  If the Directory
+#	Listing is hidden, then give the not-found page.
+#
+# Arguments:
+#	prefix	The URL domain prefix.
+#	path	The file system pathname of the directory.
+#	suffix	The URL suffix.
+#	sock	The socket connection.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Dispatches to the appropriate page handler.
+
+proc DirList_Directory {prefix path suffix sock} {
+    upvar #0 Httpd$sock data
+    global dirlist tcl_platform
+
+    # Special case because glob doesn't work in wrapped files
+    # Just set indexpat to "index.tml" or "index.html"
+
+    set npath [file join $path $dirlist(indexpat)]
+    if {[info exist tcl_platform(isWrapped)] && $tcl_platform(isWrapped)} {
+	set newest $npath
+    } else {
+	set newest [file_latest [glob -nocomplain $npath]]
+    }
+    if {[string length $newest]} {
+
+	# Template hack.  Ask for the corresponding .html file in
+	# case that file should be cached when running the template.
+	# If we ask for the .tml directly then its result is never cached.
+	global Template
+	if {[string compare $Template(tmlExt) [file extension $newest]] == 0} {
+	    set newest [file root $newest]$Template(htmlExt)
+	}
+	return [Doc_Handle $prefix $newest $suffix $sock]
+    }
+    if {[Dir_ListingIsHidden]} {
+        # Directory listings are hidden, so give the not-found page.
+        return [Doc_NotFound $sock]
+    }
+    # Listings are not hidden, so show it.
+    Httpd_ReturnData $sock text/html [DirList $sock $path $data(url)]
+}
+
+# Dir_HideListings --
+#
+#	If a directory is viewed, hide the directory listing.
+#
+# Arguments:
+#	None
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	From now on, directory listings are hidden.
+
+proc Dir_HideListings {} {
+    global Doc
+    set Doc(HideDirListing) 1
+    return
+}
+
+# Dir_ShowListings --
+#
+#	If a directory is viewed, show the directory listing.
+#
+# Arguments:
+#	None
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	From now on, directory listings can be shown.
+
+proc Dir_ShowListings {} {
+    global Doc
+    set Doc(HideDirListing) 0
+    return
+}
+
+# Dir_ListingIsHidden --
+#
+#	Tell whether directory listings are currently hidden.
+#
+# Arguments:
+#	None
+#
+# Results:
+#	Returns 1 if listings are hidden, otherwise 0.
+#
+# Side Effects:
+#	None
+
+proc Dir_ListingIsHidden {} {
+    global Doc
+    return $Doc(HideDirListing)
+}
+
+# By default, directory listings are shown.
+Dir_ShowListings
+
 proc DirListForm {dir urlpath {sort name} {pattern *}} {
     set what [DirListTerm]
     set namecheck ""
@@ -45,6 +172,11 @@ or Size <input type=radio name=sort value=size $sizecheck><br>
 proc DirListInner {dir urlpath sort pattern} {
     set listing "<PRE>\n"
     set path [file split $dir]
+
+    # Filter pattern to avoid leaking path information
+    regsub -all {\.+/} $pattern {} pattern
+    set pattern [string trimleft $pattern /]
+
     set list [glob -nocomplain -- [file join $dir $pattern]]
     if {[llength $path] > 1} {
 	append listing \
@@ -134,8 +266,7 @@ proc DirHref {entry} {
     return $entry
 }
 
-proc DirList {dir urlpath} {
-    upvar 1 sock sock	;# DISGUSTING HACK
+proc DirList {sock dir urlpath} {
     upvar #0 Httpd$sock data
 
     set sort name
