@@ -18,12 +18,13 @@
 # Matt Newman (c) 1999 Novadigm Inc.
 # Stephen Uhler / Brent Welch (c) 1997 Sun Microsystems
 # Brent Welch (c) 1998-2000 Ajuba Solutions
+# Brent Welch (c) 2001-2004 Panasas Inc
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# RCS: @(#) $Id: httpd.tcl,v 1.69.2.1 2002/08/18 22:54:17 welch Exp $
+# RCS: @(#) $Id: httpd.tcl,v 1.85 2004/04/29 01:34:16 coldstore Exp $
 
-package provide httpd 1.6
+package provide httpd 1.7
 
 # initialize all the global data
 
@@ -94,6 +95,8 @@ array set Httpd_EnvMap {
 # count		Content-Length
 # set-cookie	List of Set-Cookie headers to stick into the response
 #		Use Httpd_SetCookie to append to this.
+# headers		List of http headers to stick into the response
+#		Use Httpd_AddHeaders to append to this.
 
 # prefix	(Set by Url_Dispatch to be the URL domain prefix)
 # suffix	(Set by Url_Dispatch to be the URL domain suffix)
@@ -599,8 +602,8 @@ proc HttpdRead {sock} {
     #	mime: we are reading the protocol headers
     # and how much was read. Note that
     # [string compare $readCount 0] maps -1 to -1, 0 to 0, and > 0 to 1
-
     set state [string compare $readCount 0],$data(state)
+
     switch -glob -- $state {
 	1,start	{
 	    if {[regexp {^([^ ]+) +([^?]+)\??([^ ]*) +HTTP/(1.[01])} \
@@ -657,6 +660,15 @@ proc HttpdRead {sock} {
 	    # MIME headers.  MIME headers may be continue with a line
 	    # that starts with spaces.
 	    if {[regexp {^([^ :]+):[ 	]*(.*)} $line dummy key value]} {
+
+                # The following allows something to
+                # recreate the headers exactly
+
+                lappend data(headerlist) $key $value
+
+                # The rest of this makes it easier to pick out
+                # headers from the data(mime,headername) array
+
 		set key [string tolower $key]
 		if [info exists data(mime,$key)] {
 		    append data(mime,$key) ,$value
@@ -665,6 +677,7 @@ proc HttpdRead {sock} {
 		    lappend data(mimeorder) $key
 		}
 		set data(key) $key
+
 	    } elseif {[regexp {^[ 	]+(.*)}  $line dummy value]} {
 		# Are there really continuation lines in the spec?
 		if [info exists data(key)] {
@@ -694,7 +707,7 @@ proc HttpdRead {sock} {
 	    unset data
 	    $i eval [list fileevent $sock readable [list HttpdRead $sock]]
 	    set tmp [$i eval [list \
-		    after {$Httpd(timeout2)} [list HttpdCancel $sock]]]
+		    after $Httpd(timeout2) [list HttpdCancel $sock]]]
 	    $i eval [list array set Httpd$sock [list cancel $tmp]]
 	}
 	0,mime	{
@@ -727,6 +740,7 @@ proc HttpdRead {sock} {
 
 		Url_PostHook $sock $data(count)
 	    } else {
+		Url_PostHook $sock 0    ;# Clear any left-over hook
 		set data(count) 0
 	    }
 
@@ -1105,6 +1119,74 @@ proc HttpdDoCallback {sock {errmsg {}}} {
     CountStop serviceTime $sock
 }
 
+# Httpd_AddHeaders
+#	Add http headers to be used in a reply
+#	Call this before using Httpd_ReturnFile or
+#	Httpd_ReturnData
+#
+# Arguments:
+#	sock	handle on the connection
+#	args	a list of header value ...
+
+proc Httpd_AddHeaders {sock args} {
+    upvar #0 Httpd$sock data
+
+    eval lappend data(headers) $args
+}
+
+# Httpd_RemoveHeaders
+#	Remove previously set headers from the reply.
+#	Any headers that match the glob pattern are removed.
+#
+# Arguments:
+#	sock	handle on the connection
+#	pattern	glob pattern to match agains cookies.
+
+proc Httpd_RemoveHeaders {sock {pattern *}} {
+    upvar #0 Httpd$sock data
+    if {[info exists data(headers)] && $data(headers) != {}} {
+        set tmp {}
+        foreach {header value} $data(headers) {
+            if {![string match $pattern $header]} {
+                lappend tmp $header $value
+            }
+        }
+        set data(headers) $tmp
+    }
+    return
+}
+
+# Httpd_NoCache
+#	Insert header into http header to indicate that this page
+#	should not be cached
+#
+# Arguments:
+#	sock	handle on the connection
+
+proc Httpd_NoCache {sock} {
+    Httpd_RemoveHeaders $sock Cache-Control
+    Httpd_AddHeaders $sock Cache-Control no-cache
+    Httpd_AddHeaders $sock Expires content '-1'
+}
+
+# Httpd_Refresh
+#	Insert header into http header to cause browser to refresh
+#	after a delay, optionally to a different URL
+#
+# Arguments:
+#	sock	handle on the connection
+#	time	time in seconds before refresh
+#	url	optional: url to refresh to
+
+proc Httpd_Refresh {sock time {url ""}} {
+    Httpd_RemoveHeaders $sock Cache-Control
+    if {$url == ""} {
+	Httpd_AddHeaders $sock Refresh $time
+    } else {
+	Httpd_AddHeaders $sock Refresh ${time}\;url=${url}
+    }
+}
+
 # HttpdRespondHeader --
 #
 #	Utility routine for outputting response headers for normal data Does
@@ -1139,6 +1221,15 @@ proc HttpdRespondHeader {sock type close size {code 200}} {
     if {[string length $size]} {
 	append reply "Content-Length: $size" \n
     }
+
+    if {[info exists data(headers)]} {
+	foreach {header value} $data(headers) {
+	    catch {
+		append reply "[string trimright $header :]: " $value \n
+	    }
+	}
+    }
+
     puts -nonewline $sock $reply
 }
 
@@ -1229,6 +1320,7 @@ proc HttpdSetCookie {sock} {
 #	sock	handle on the connection
 #	type	is a Content-Type
 #	path	is the file pathname
+#	offset	amount to skip at the start of file
 #
 # Side Effects:
 #	Sends the file contents back as the reply.
@@ -1242,6 +1334,11 @@ proc Httpd_ReturnFile {sock type path {offset 0}} {
 	return
     }
 
+    # Set file size early so it gets into all log records
+
+    set data(file_size) [file size $path]
+    set data(code) 200
+
     Count urlreply
     if {[info exists data(mime,if-modified-since)]} {
         # No need for complicated date comparison, if they're identical then 304.
@@ -1250,7 +1347,6 @@ proc Httpd_ReturnFile {sock type path {offset 0}} {
             return
         }
     } 
-    set data(file_size) [file size $path]
 
     # Some files have a duality, when the client sees X bytes but the
     # file is really X + n bytes (the first n bytes reserved for server
@@ -1286,6 +1382,7 @@ proc Httpd_ReturnFile {sock type path {offset 0}} {
 #	Return data for a page.
 #
 # Arguments:
+#	sock	handle on the connection
 #	type	a Content-Type
 #	content	the data to return
 #	code	the HTTP reply code.
@@ -1361,8 +1458,6 @@ proc Httpd_ReturnCacheableData {sock type content date {code 200}} {
 }
 
 # HttpdCopyDone -- this is used with fcopy when the copy completes.
-# Note that tcl8.0b1 had a bug in that errors during fcopy called
-# bgerror instead of this routine, which causes leaks.  Don't use b1.
 #
 # Arguments:
 #	in	Input channel, typically a file
@@ -1378,6 +1473,11 @@ proc Httpd_ReturnCacheableData {sock type content date {code 200}} {
 #	See Httpd_SockClose
 
 proc HttpdCopyDone {in sock close bytes {error {}}} {
+    if {$error == ""} {
+        # This special value signals a normal close,
+        # and triggers a log record so static files are counted
+        set error Close
+    }
     Httpd_SockClose $sock $close $error
 }
 
@@ -1442,7 +1542,7 @@ proc Httpd_Error {sock code {detail ""}} {
     if {$code == 500} {
 	append message "<h2>Tcl Call Trace</h2>"
 	for {set l [expr [info level]-1]} {$l > 0} {incr l -1} {
-		append message "$l: [info level $l]<br>"
+	    append message "$l: [protect_text [info level $l]]<br>"
 	}
     }
     Log $sock Error $code $data(url) $detail
@@ -1459,7 +1559,7 @@ proc Httpd_Error {sock code {detail ""}} {
 	return
     }
     if [catch {
-	HttpdRespondHeader $sock text/html 1 [string length $message] $code
+	HttpdRespondHeader $sock text/html 1 [expr {[string length $message] + 4}] $code
 	puts $sock ""
 	puts $sock $message
     } err] {
@@ -1756,6 +1856,7 @@ set HttpdAuthorizationFormat {
 #	sock	Socket connection
 #	type	usually "Basic"
 #	realm	browsers use this to cache credentials
+#	args	additional name value pairs for request
 #
 # Results:
 #	None
@@ -1763,7 +1864,7 @@ set HttpdAuthorizationFormat {
 # Side Effects:
 #	Generate an authorization challenge response.
 
-proc Httpd_RequestAuth {sock type realm} {
+proc Httpd_RequestAuth {sock type realm args} {
     upvar #0 Httpd$sock data
     global Httpd HttpdAuthorizationFormat
 
@@ -1772,9 +1873,14 @@ proc Httpd_RequestAuth {sock type realm} {
 	return
     }
 
+    set additional ""
+    foreach {name value} $args {
+	append additional ", " ${name}=$value
+    }
+
     set close [HttpdCloseP $sock]
     HttpdRespondHeader $sock text/html $close [string length $HttpdAuthorizationFormat] 401
-    puts $sock "Www-Authenticate: $type realm=\"$realm\""
+    puts $sock "Www-Authenticate: $type realm=\"$realm\" $additional"
     puts $sock ""
     puts -nonewline $sock $HttpdAuthorizationFormat
     Httpd_SockClose $sock $close
@@ -1953,7 +2059,7 @@ proc Httpd_RequestComplete {sock} {
 
 	set msg "[clock format [clock seconds]]\n$errorInfo"
 	if [catch {Log nosock bgerror $msg}] {
-	    catch {puts stderr $msg}
+	    Stderr $msg
 	}
     }
 
@@ -2002,6 +2108,7 @@ proc HttpdCookieLog {sock what} {
 #
 #	Suspend Wire Callback - for async transactions
 #	Use Httpd_Resume once you are back in business
+#	Note: global page array is not preserved over suspend
 #
 # Arguments:
 #	sock	Socket connection
@@ -2186,3 +2293,33 @@ proc Httpd_DumpHeaders {sock} {
     }
     return $result
 }
+
+# Httpd_Webmaster --
+#
+# Define an email address for the webmaster
+#
+# Arguments:
+#	email 	The email of the webmaster.  If empty, the
+#		current value is returned, which is handy in
+#		web pages.
+#
+# Results:
+#	Returns the webmaster email.
+#
+# Side Effects:
+#	Sets the webmaster email.
+
+proc Httpd_Webmaster {{email {}}} {
+    global Httpd
+    if {[string length $email] == 0} {
+	if {![info exists Httpd(webmaster)]} {
+	    set Httpd(webmaster) webmaster
+	}
+	return $Httpd(webmaster)
+    } else {
+	set Httpd(webmaster) $email
+    }
+}
+
+# this is too much of a compatibility hassle, so we alias
+catch {interp alias {} Doc_Webmaster {} Httpd_Webmaster}
